""" Dataframe functions """
import logging
import os
from tempfile import mkstemp

import pandas as pd
from box import Box

# pylint: disable=too-many-arguments

logger = logging.getLogger(__name__)  # pylint: disable=C0103


def pd_export(
    dataframe: pd.DataFrame,
    export_type: str,
    df_name: str,
    temp_name: bool = False,
    df_name_prefix: str = "",
    df_name_suffix: str = "",
    dir_name: str = ".",
    config_box: Box = None,
    index=True,
    header=True,
) -> str:
    """
    Exports dataframe to file formats using various options

    Return a filepaths for the exported Dataframe
    """

    if temp_name and dir_name != "":
        filepath = mkstemp(suffix=df_name_suffix, prefix=df_name_prefix, dir=dir_name)[
            1
        ]

    elif config_box and dir_name == "":
        filepath = os.path.join(
            config_box.extracttempdir,
            f"{df_name_prefix}{df_name}{df_name_suffix}.{export_type}",
        )
    else:
        filename = f"{df_name_prefix}{df_name}{df_name_suffix}.{export_type}"
        filepath = os.path.join(dir_name, filename)

    logger.info("Creating %s file %s from dataframe.", export_type, filepath)

    if export_type == "parquet":
        dataframe.to_parquet(path=filepath, index=index)
    elif export_type == "csv":
        dataframe.to_csv(filepath, index=index, header=header)

    return filepath


def pd_colupdate(dataframe: pd.DataFrame, coldict: dict) -> pd.DataFrame:
    """
    Rename and filter Pandas Dataframe columns using python dictionary.

    Column names provided in coldict follow the same format as expected by
    pd.DataFrame.rename(columns=dict). For example: {"current":"new", "current2":"new2"}

    Columns in returned dataframe are filtered by those provided to be renamed.

    Returns a modified pd.Dataframe copy
    """

    logger.info("Renaming and filtering dataframe columns using coldict key:values.")

    # Remap column names
    dataframe = dataframe.rename(columns=coldict)

    # Filter columns based on the new names
    dataframe = dataframe[[val for key, val in coldict.items()]].copy()

    return dataframe



""" SFTP related Prefect tasks """
import datetime
import os
from typing import Any

import pandas as pd
import prefect
import pysftp
from box import Box
from prefect import Task

# pylint: disable=arguments-differ, no-member, logging-too-many-args


class SFTPExists(Task):
    """
    Checks filename from FTP server

    Return True if found, otherwise False
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def run(
        self,
        workfile: str,
        config_box: Box,
        cnopts: pysftp.CnOpts = None,
        **format_kwargs: Any,
    ) -> str:
        with prefect.context(**format_kwargs) as data:

            if data.get("parameters"):
                if data.parameters.get("cnopts"):
                    cnopts = data.parameters["cnopts"]

            # We have to handle either a password or a key
            if config_box.get("private_key_path"):
                private_key = config_box["private_key_path"]
                if config_box.get("private_key_passphrase"):
                    # has a passphrase, use it
                    private_key_passphrase = config_box["private_key_passphrase"]
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        private_key_pass=private_key_passphrase,
                        cnopts=cnopts,
                    )
                else:
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        cnopts=cnopts,
                    )
            elif config_box.get("password"):
                password = config_box["password"]
                sftp = pysftp.Connection(
                    host=config_box["hostname"],
                    username=config_box["username"],
                    password=password,
                    cnopts=cnopts,
                )

            try:
                with sftp.cd(config_box["target_dir"]):
                    result = sftp.exists(workfile)
            finally:
                sftp.close()

            return result


class SFTPGet(Task):
    """
    Fetch filename from FTP server

    Return a file_location_name
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def run(
        self,
        workfile: str,
        config_box: Box,
        cnopts: pysftp.CnOpts = None,
        tempfolderpath: str = None,
        **format_kwargs: Any,
    ) -> str:
        with prefect.context(**format_kwargs) as data:

            if data.get("parameters"):
                if data.parameters.get("cnopts"):
                    cnopts = data.parameters["cnopts"]
                if data.parameters.get("cache"):
                    tempfolderpath = data.parameters["cache"]

            localtmpfile = os.path.join(tempfolderpath, workfile)
            self.logger.debug("Working on {}", os.path.join(tempfolderpath, workfile))

            if config_box.get("private_key_path"):
                private_key = config_box["private_key_path"]
                if config_box.get("private_key_passphrase"):
                    # has a passphrase, use it
                    private_key_passphrase = config_box["private_key_passphrase"]
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        private_key_pass=private_key_passphrase,
                        cnopts=cnopts,
                    )
                else:
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        cnopts=cnopts,
                    )
            elif config_box.get("password"):
                password = config_box["password"]
                sftp = pysftp.Connection(
                    host=config_box["hostname"],
                    username=config_box["username"],
                    password=password,
                    cnopts=cnopts,
                )

            try:
                with sftp.cd(config_box["target_dir"]):
                    sftp.get(workfile, localpath=localtmpfile, preserve_mtime=False)
            finally:
                sftp.close()

            self.logger.info("SFTPGet {}", localtmpfile)

            return localtmpfile


class SFTPPut(Task):
    """
    Put a file on the FTP server

    Leave remotepath off, or None and the workfile and the remote file are the same.
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def run(
        self,
        workfile: str,
        config_box: Box,
        cnopts: pysftp.CnOpts = None,
        remotepath: str = None,
        **format_kwargs: Any,
    ):
        with prefect.context(**format_kwargs) as data:

            if data.get("parameters"):
                if data.parameters.get("cnopts"):
                    cnopts = data.parameters["cnopts"]

            if config_box.get("private_key_path"):
                private_key = config_box["private_key_path"]
                if config_box.get("private_key_passphrase"):
                    # has a passphrase, use it
                    private_key_passphrase = config_box["private_key_passphrase"]
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        private_key_pass=private_key_passphrase,
                        cnopts=cnopts,
                    )
                else:
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        cnopts=cnopts,
                    )
            elif config_box.get("password"):
                password = config_box["password"]
                sftp = pysftp.Connection(
                    host=config_box["hostname"],
                    username=config_box["username"],
                    password=password,
                    cnopts=cnopts,
                )

            try:
                if not sftp.isdir(config_box["target_dir"]):
                    sftp.mkdir(config_box["target_dir"])

                with sftp.cd(config_box["target_dir"]):
                    sftp.put(workfile, preserve_mtime=False, remotepath=remotepath)
            finally:
                sftp.close()

            self.logger.info("SFTPPut {}", workfile)


class SFTPRemove(Task):
    """
    Remove file from the FTP server
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def run(
        self,
        workfile: str,
        config_box: Box,
        cnopts: pysftp.CnOpts = None,
        **format_kwargs: Any,
    ):
        with prefect.context(**format_kwargs) as data:

            self.logger.debug("Attempting to remove ", workfile)

            if data.get("parameters"):
                if data.parameters.get("cnopts"):
                    cnopts = data.parameters["cnopts"]

            if config_box.get("private_key_path"):
                private_key = config_box["private_key_path"]
                if config_box.get("private_key_passphrase"):
                    # has a passphrase, use it
                    private_key_passphrase = config_box["private_key_passphrase"]
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        private_key_pass=private_key_passphrase,
                        cnopts=cnopts,
                    )
                else:
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        cnopts=cnopts,
                    )
            elif config_box.get("password"):
                password = config_box["password"]
                sftp = pysftp.Connection(
                    host=config_box["hostname"],
                    username=config_box["username"],
                    password=password,
                    cnopts=cnopts,
                )

            # Pick out the oldest file in the dataframe
            try:
                with sftp.cd(config_box["target_dir"]):
                    sftp.remove(workfile)
            finally:
                sftp.close()

            # Read the file into a dataframe
            self.logger.info("SFTPRemove {}", workfile)


class SFTPRename(Task):
    """
    Rename (move) file on the FTP server
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def run(
        self,
        config_box: Box,
        source="root",
        target="wip",
        cnopts: pysftp.CnOpts = None,
        **format_kwargs: Any,
    ):
        with prefect.context(**format_kwargs) as data:

            # This moves (renames) files around on the SFTP site.
            # It's helpful for managing "state" as the files are processed locally.

            if data.parameters.get("cnopts"):
                cnopts = data.parameters["cnopts"]

            if config_box.get("private_key_path"):
                private_key = config_box["private_key_path"]
                if config_box.get("private_key_passphrase"):
                    # has a passphrase, use it
                    private_key_passphrase = config_box["private_key_passphrase"]
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        private_key_pass=private_key_passphrase,
                        cnopts=cnopts,
                    )
                else:
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        cnopts=cnopts,
                    )
            elif config_box.get("password"):
                password = config_box["password"]
                sftp = pysftp.Connection(
                    host=config_box["hostname"],
                    username=config_box["username"],
                    password=password,
                    cnopts=cnopts,
                )

            # "root" is special
            remotesourcepath = (
                config_box["target_dir"]
                if source != "root"
                else os.path.join(config_box["target_dir"], source)
            )

            remotetargetpath = os.path.join(config_box["target_dir"], target)
            filename = os.path.basename(remotesourcepath)

            self.logger.debug(
                "Moving {} from {} to {}", filename, remotesourcepath, remotetargetpath
            )

            # Move the file from source to target on the SFTP
            try:
                with sftp.cd(config_box["target_dir"]):

                    if not sftp.isfile(os.path.join(remotesourcepath, filename)):
                        # Working in notebooks you might have already
                        # moved the file in another block
                        self.logger.warning(
                            "The file {} isn't in the remote folder.",
                            os.path.join(remotesourcepath, filename),
                        )
                    else:
                        if not sftp.isdir(remotetargetpath):
                            sftp.mkdir(remotetargetpath)

                        if not sftp.isfile(os.path.join(remotetargetpath, filename)):
                            # The file might already be in DONE...
                            sftp.rename(
                                os.path.join(remotesourcepath, filename),
                                os.path.join(remotetargetpath, filename),
                            )
            finally:
                sftp.close()


class SFTPPoll(Task):
    """
    Polls for SFTP files
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def run(
        self, config_box: Box, cnopts: pysftp.CnOpts = None, **format_kwargs: Any
    ) -> pd.DataFrame:
        with prefect.context(**format_kwargs) as data:

            if data.get("parameters"):
                if data.parameters.get("cnopts"):
                    cnopts = data.parameters["cnopts"]

            if config_box.get("private_key_path"):
                private_key = config_box["private_key_path"]
                if config_box.get("private_key_passphrase"):
                    # has a passphrase, use it
                    private_key_passphrase = config_box["private_key_passphrase"]
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        private_key_pass=private_key_passphrase,
                        cnopts=cnopts,
                    )
                else:
                    sftp = pysftp.Connection(
                        host=config_box["hostname"],
                        username=config_box["username"],
                        private_key=private_key,
                        cnopts=cnopts,
                    )
            elif config_box.get("password"):
                password = config_box["password"]
                sftp = pysftp.Connection(
                    host=config_box["hostname"],
                    username=config_box["username"],
                    password=password,
                    cnopts=cnopts,
                )

            files_data = []

            try:
                with sftp.cd(config_box["target_dir"]):
                    for dir_file in sftp.listdir():

                        # Extra dirs like wip and done require us to
                        # check if we're just looking at files
                        if sftp.isfile(dir_file):

                            # get the dates from the ftp site itself
                            sftpattrs = sftp.stat(dir_file)

                            files_data.append(
                                {
                                    "File Name": dir_file,
                                    "MTime": datetime.datetime.fromtimestamp(
                                        sftpattrs.st_mtime
                                    ),
                                }
                            )
            finally:
                sftp.close()

            files_df = pd.DataFrame(files_data, columns=["File Name", "MTime"])

            self.logger.info("Found {} files to process.", len(files_df.index))
            return files_df


class DFGetOldestFile(Task):
    """
    Pick the oldest file off the top of the given dataframe.
    The dataframe requires columns called 'File Name' and 'MTime'

    Includes a search string to filter the list

    Returns a filename to fetch.
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def run(self, files_df: pd.DataFrame, regex_search: str) -> str:

        if "MTime" not in files_df:
            raise ValueError("The MTime column is missing from the dataframe.")

        if "File Name" not in files_df:
            raise ValueError("The 'File Name' column is missing from the dataframe.")

        if len(files_df.index) == 0:
            self.logger.debug("The given DataFrame is empty.")
            return None

        if regex_search:
            files_df = files_df[
                files_df["File Name"].str.contains(regex_search, regex=True)
            ]

        files_df = files_df.sort_values(by="MTime")

        # Fetch the oldest file from the frame and bring it to a local temp file.
        workfile = files_df.iloc[0]["File Name"]
        self.logger.info("Found oldest file, {}", workfile)
        return workfile



""" Factory for creating tasks from nuggets """
from collections.abc import Callable

import prefect
from . import nuggets

# prepare a placeholder class for prefect tasks created from nuggets
ptask = lambda: None  # pylint: disable=C0103

# gather list of nuggets
NUGGET_LIST = [
    f
    for f in dir(nuggets)
    # exclude built-ins and include only functions
    if not f.startswith("__") and isinstance(getattr(nuggets, f), Callable)
]

# for each nugget, add it (and a parent) attr to ptask
for nugget in NUGGET_LIST:

    # parent attr for task
    parent = getattr(nuggets, nugget).__module__[20:]

    # if no parent attr attached to ptask, set it
    if not hasattr(ptask, parent):
        setattr(
            ptask,
            parent,
            lambda: None,
        )

    # set attr per parent attr within ptask
    setattr(
        getattr(ptask, parent),
        getattr(nuggets, nugget).__name__,
        prefect.task(getattr(nuggets, nugget)),
    )



""" Validation functions """
import logging

import fastavro as avro
import pandas as pd
import pyarrow as pa

logger = logging.getLogger(__name__)  # pylint: disable=C0103


def df_arrow_validate(dataframe: pd.DataFrame, arsc: pa.lib.Schema) -> bool:
    """Validate Pandas dataframe against arrow schema"""
    logger.info("Validating dataframe against arrow schema")
    return arsc.equals(pa.Schema.from_pandas(dataframe))


def df_avro_validate(dataframe: pd.DataFrame, avsc: dict) -> bool:
    """Validate Pandas dataframe against avro schema dict"""
    logger.info("Validating dataframe against avro schema dict.")
    return avro.validation.validate_many(
        records=dataframe.replace(pd.NA, "").to_dict(orient="records"), schema=avsc
    )



""" Init for cupyopt base """
from .objectstore_tasks import (
    ObjstrClient,
    ObjstrFGet,
    ObjstrFPut,
    ObjstrGet,
    ObjstrGetAsDF,
    ObjstrMakeBucket,
    ObjstrPut,
)
from .oradb_tasks import ORADBGetEngine, ORADBSelectToDataFrame
from .sftp_tasks import (
    DFGetOldestFile,
    SFTPExists,
    SFTPGet,
    SFTPPoll,
    SFTPPut,
    SFTPRemove,
)
from .task_factory import ptask



""" Init for cupyopt nuggets """
from .dataframe import pd_colupdate, pd_export
from .schema import (
    arrow_schema_from_parquet,
    arrow_schema_to_parquet,
    avro_schema,
    avro_schema_to_file,
    infer_df_arrow_schema,
    infer_df_avro_schema,
)
from .validation import df_arrow_validate, df_avro_validate



""" Schema functions """
import json
import logging
from typing import Union

import fastavro as avro
import pandas as pd
import pandavro as pda
import pyarrow as pa
import pyarrow.parquet as pq

logger = logging.getLogger(__name__)  # pylint: disable=C0103


def infer_df_arrow_schema(dataframe: pd.DataFrame) -> pa.lib.Schema:
    """Infer arrow schema from pandas dataframe"""
    logging.info("Inferring arrow schema from dataframe")
    # infer the schema using pyarrow
    arsc = pa.Schema.from_pandas(df=dataframe)

    return arsc


def arrow_schema_to_parquet(
    arsc: pa.lib.Schema, filename: str, filedir: str = "."
) -> str:
    """Export arrow schema to parquet file"""
    logging.info("Exporting arrow schema to parquet file")

    filepath = f"{filedir}/{filename}.schema.parquet"

    # write schema into empty table parquet file
    pq.write_table(table=arsc.empty_table(), where=filepath)

    return filepath


def arrow_schema_from_parquet(source: Union[pa.lib.Table, str]) -> pa.lib.Schema:
    """Import arrow schema from parquet file"""
    logging.info("Importing arrow schema from parquet")
    if isinstance(source, pa.lib.Table):
        schema = source.schema

    elif isinstance(source, str):
        schema = pq.read_table(source=source).schema

    return schema


def avro_schema(avsc: Union[dict, str]) -> dict:
    """Create avro schema from dictionary or filepath string"""
    # if a dictionary type, parse from dict
    logging.info("Parsing avro schema")
    if isinstance(avsc, dict):
        avsc = avro.schema.parse_schema(avsc)

    # if a str type, load from file
    elif isinstance(avsc, str):
        avsc = avro.schema.load_schema(avsc)

    return avsc


def infer_df_avro_schema(
    dataframe: pd.DataFrame,
    name: str = None,
    namespace: str = None,
    times_as_micros: bool = True,
) -> dict:
    """Infer avro schema from pandas dataframe"""
    logging.info("Inferring avro schema from dataframe")
    # infer the schema using pandavro
    schema = pda.schema_infer(df=dataframe, times_as_micros=times_as_micros)

    # add custom schema name if exists (by default "Root")
    if name:
        schema["name"] = name
    # add custom schema name if exists (by default, non-existent)
    if namespace:
        schema["namespace"] = namespace

    return avro_schema(schema)


def avro_schema_to_file(avsc: dict, filename: str = None, filedir: str = ".") -> str:
    """Export avro schema to file"""
    logging.info("Exporting avro schema to file")
    # infer the filename based on the avro schema name from the avro dict key:values
    if not filename:
        filename = avsc["name"]

    # create internal copy of avro dict so as to not otherwise interfere
    _avsc = avsc.copy()

    # remove additional keys fastavro places in dict
    for key in ["__named_schemas", "__fastavro_parsed"]:
        if key in _avsc.keys():
            _avsc.pop(key)

    filepath = f"{filedir}/{filename}.avsc"

    with open(filepath, "w", encoding="utf-8") as avro_file:
        avro_file.write(json.dumps(_avsc, indent=4))

    return filepath



""" object store functions """

import logging
from typing import Any
from typing_extensions import Literal

import pandas as pd
import urllib3
from box import Box
from minio import Minio
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

# pylint: disable=arguments-differ, too-many-arguments


class ObjstrClient(Task):
    """setup object storage client via minio"""

    def __init__(
        self,
        config_box: Box = None,
        http_client: urllib3.poolmanager.PoolManager = None,
        **kwargs: Any
    ):
        self.config_box = config_box
        self.http_client = http_client
        super().__init__(**kwargs)

    @defaults_from_attrs("config_box", "http_client")
    def run(
        self, config_box: Box, http_client: urllib3.poolmanager.PoolManager = None
    ) -> Minio:

        logging.info(
            "Creating object store client for endpoint %s.", config_box.endpoint
        )
        client = Minio(
            endpoint=config_box.endpoint,
            access_key=config_box.key,
            secret_key=config_box.secret,
            secure=True,
            http_client=http_client,
        )

        return client


class ObjstrMakeBucket(Task):
    """attempts to make a bucket if not already available"""

    def __init__(self, client: Minio = None, bucket_name: str = None, **kwargs: Any):
        self.client = client
        self.bucket_name = bucket_name

        super().__init__(**kwargs)

    @defaults_from_attrs("client", "bucket_name")
    def run(self, client: Minio, bucket_name: str):

        # make bucket if not exists
        if not client.bucket_exists(bucket_name):
            client.make_bucket(bucket_name)
            logging.info("Created bucket %s", bucket_name)
        else:
            logging.info("Bucket %s already exists, taking no actions.", bucket_name)


class ObjstrPut(Task):
    """put data as object in object store"""

    def __init__(
        self,
        client: Minio = None,
        bucket_name: str = None,
        object_name: str = None,
        data: object = None,
        length: int = -1,
        part_size: int = 5 * 1024 * 1024,
        **kwargs: Any
    ):

        self.client = client
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.data = data
        self.length = length
        self.part_size = part_size

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "bucket_name", "object_name", "data", "length", "part_size", "client"
    )
    def run(
        self,
        client: Minio,
        bucket_name: str,
        object_name: str,
        data: object,
        length: int = None,
        part_size: int = None,
    ):
        if not length:
            length = self.length

        if not part_size:
            part_size = self.part_size

        # upload file as object
        client.put_object(
            bucket_name=bucket_name,
            object_name=object_name,
            data=data,
            length=length,
            part_size=part_size,
        )

        logging.info("Put data under %s as %s", bucket_name, object_name)


class ObjstrGet(Task):
    """get object as data from object store"""

    def __init__(
        self,
        client: Minio = None,
        bucket_name: str = None,
        object_name: str = None,
        **kwargs: Any
    ):
        self.client = client
        self.bucket_name = bucket_name
        self.object_name = object_name

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "client",
        "bucket_name",
        "object_name",
    )
    def run(self, client: Minio, bucket_name: str, object_name: str) -> Any:

        # get object as file
        data = client.get_object(
            bucket_name=bucket_name,
            object_name=object_name,
        )

        logging.info(
            "Retrieved object %s under %s",
            object_name,
            bucket_name,
        )
        return data


class ObjstrGetAsDF(Task):
    """get object and return as pandas.dataframe from object store"""

    def __init__(
        self,
        client: Minio = None,
        bucket_name: str = None,
        object_name: str = None,
        dftype: Literal["csv", "parquet", "excel"] = None,
        **kwargs: Any
    ):
        self.client = client
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.dftype = dftype

        super().__init__(**kwargs)

    @defaults_from_attrs("client", "bucket_name", "object_name", "dftype")
    def run(
        self,
        client: Minio,
        bucket_name: str,
        object_name: str,
        dftype: Literal["csv", "parquet", "excel"],
        **kwargs: Any
    ) -> Any:

        # get object as file
        data = client.get_object(
            bucket_name=bucket_name,
            object_name=object_name,
        )

        logging.info(
            "Retrieved object %s under %s",
            object_name,
            bucket_name,
        )

        if dftype == "csv":
            pd_dataframe = pd.read_csv(filepath_or_buffer=data, **kwargs)
        elif dftype == "parquet":
            pd_dataframe = pd.read_parquet(filepath_or_buffer=data, **kwargs)

        return pd_dataframe


class ObjstrFPut(Task):
    """put file as object in object store"""

    def __init__(
        self,
        client: Minio = None,
        bucket_name: str = None,
        object_name: str = None,
        file_path: str = None,
        **kwargs: Any
    ):

        self.client = client
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.file_path = file_path

        super().__init__(**kwargs)

    @defaults_from_attrs("client", "bucket_name", "object_name", "file_path")
    def run(
        self,
        client: Minio,
        bucket_name: str,
        object_name: str,
        file_path: str,
    ):

        # upload file as object
        client.fput_object(
            bucket_name=bucket_name,
            object_name=object_name,
            file_path=file_path,
        )

        logging.info("Put file %s under %s as %s", file_path, bucket_name, object_name)


class ObjstrFGet(Task):
    """get object as file from object store"""

    def __init__(
        self,
        client: Minio = None,
        bucket_name: str = None,
        object_name: str = None,
        file_path: str = None,
        **kwargs: Any
    ):
        self.client = client
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.file_path = file_path

        super().__init__(**kwargs)

    @defaults_from_attrs("client", "bucket_name", "object_name", "file_path")
    def run(self, client: Minio, bucket_name: str, object_name: str, file_path: str):

        # get object as file
        client.fget_object(
            bucket_name=bucket_name,
            object_name=object_name,
            file_path=file_path,
        )

        logging.info(
            "Retrieved object %s under %s as file %s",
            object_name,
            bucket_name,
            file_path,
        )



""" Setup for cupyopt """
import os
import pathlib

import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.rst").read_text()

# Pull requirements from the text file
REQUIREMENT_PATH = HERE / "requirements.txt"
INSTALL_REQUIRES = []
if os.path.isfile(REQUIREMENT_PATH):
    with open(REQUIREMENT_PATH) as f:
        INSTALL_REQUIRES = f.read().splitlines()

# This call to setup() does all the work
setuptools.setup(
    name="cupyopt",
    version="0.15.11.2",
    description="CU Python Opinionated Prefect Tasks",
    long_description=README,
    long_description_content_type="text/x-rst",
    author="CU Boulder, OIT",
    author_email="stta9820@colorado.edu",
    license="Apache License 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    package_dir={"": "src"},
    install_requires=INSTALL_REQUIRES,
)



""" oracle database related Prefect tasks """

import os
from typing import Any

import pandas as pd
import sqlalchemy
from box import Box
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs
from sqlalchemy import create_engine

# pylint: disable=arguments-differ


class ORADBGetEngine(Task):
    """
    Configure an sqlalchemy engine with cx_Oracle

    Return a sqlalchemy engine
    """

    def __init__(self, config_box: Box = None, is_sid: bool = None, **kwargs: Any):
        self.config_box = config_box
        self.is_sid = is_sid
        super().__init__(**kwargs)

    @defaults_from_attrs("config_box", "is_sid")
    def run(self, config_box: Box = None, is_sid: bool = None) -> str:

        if not isinstance(config_box, Box):
            raise ValueError("The configuration object must be a Box")

        # This will help encoding since most of our dbs are UTF8
        os.environ["NLS_LANG"] = ".AL32UTF8"

        self.logger.info(
            "Creating DB engine for %s@%s",
            config_box["database"],
            config_box["hostname"],
        )

        if config_box.port:
            port = config_box["port"]
        else:
            port = "1521"

        if is_sid:
            database = config_box["database"]
        else:
            database = f'?service_name={config_box["database"]}'

        if config_box.get("proxy_username"):
            username = f'{config_box["username"]}[{config_box["proxy_username"]}]'
        else:
            username = config_box["username"]

        # oracle connection string which will have
        # values replaced based on credentials provided from above json file
        oracle_connection_string = (
            "oracle+cx_oracle://{username}:{password}@{hostname}:{port}/{database}"
        )

        oracle_connection_string = oracle_connection_string.format(
            username=username,
            password=config_box["password"],
            hostname=config_box["hostname"],
            port=port,
            database=database,
        )

        # create the database connection engine using sqlalchemy library and
        # formated oracle_connection_string. Local SID instances can't support
        # identifiers longer than 128.
        if is_sid:
            engine = create_engine(oracle_connection_string, max_identifier_length=128)
        else:
            engine = create_engine(oracle_connection_string)

        return engine


class ORADBSelectToDataFrame(Task):
    """
    Runs select statement against database using SQLAlchemy engine.

    Return a Pandas DataFrame with data collected from SQL statement.
    """

    def __init__(
        self,
        select_stmt: str = None,
        engine: sqlalchemy.engine.base.Engine = None,
        **kwargs: Any,
    ):
        self.select_stmt = select_stmt
        self.engine = engine
        super().__init__(**kwargs)

    @defaults_from_attrs("select_stmt", "engine")
    def run(
        self,
        select_stmt: str = None,
        engine: sqlalchemy.engine.base.Engine = None,
    ) -> pd.DataFrame:

        self.logger.info(
            "Running select statement using SQLAlchemy engine to Pandas DataFrame."
        )

        dataframe = pd.read_sql(sql=select_stmt, con=engine)

        return dataframe
